-----------------------------------------------------------------------------------------------
-- Client Lua Script for GroupRadar
-- Copyright (c) NCsoft. All rights reserved

-- Credits:
-- arrow code, form and sprite are based on "NavMate" from Daihenka
-- overhead display function was provided by PingZing
-----------------------------------------------------------------------------------------------
 
require "Apollo"
require "Window"
require "GameLib"
require "GroupLib"
require "math"
require "string"
require "table"

-----------------------------------------------------------------------------------------------
-- GroupRadar Module Definition
-----------------------------------------------------------------------------------------------
local GroupRadar = {} 

local nLastTime, nUpdateTime, nRefreshTime = 0, 0, 0 -- time since last update and refresh
local nDefaultOffsetLeft, nDefaultOffsetTop, nDefaultOffsetRight, nDefaultOffsetBottom = -75, -100, 75, 50
local bTimed = false
local bUpdateForced = false -- set to force an immediate update, obviously you don't want to do this very often
 
-----------------------------------------------------------------------------------------------
-- "Constants"
-----------------------------------------------------------------------------------------------
local UPDATETHROTTLE, REFRESHTHROTTLE = 0.1, 0.016 -- max 10 updates and/or 60 refreshs per second

-----------------------------------------------------------------------------------------------
-- Initialization
-----------------------------------------------------------------------------------------------
function GroupRadar:new(o)
    o = o or {}
    setmetatable(o, self)
    self.__index = self 

	self.aArrowStyles = {	
		{sprite = "GroupRadar_Sprites:GroupRadar_sprArrow", bgcolor = "xkcdGreenishTurquoise", degreesOffset = 0},
		{sprite = "CRB_CharacterCreateSprites:sprCharC_GlowOrangeArrowNoBorder", bgcolor = "white", degreesOffset = 90},
		{sprite = "CRB_CharacterCreateSprites:sprCharC_GlowOrangeArrowNoBorder", bgcolor = "xkcdGreenishTurquoise", degreesOffset = 90},
		{sprite = "CRB_ChallengeSprites:sprCh_WheelSlice1", bgcolor = "xkcdGreenishTurquoise", degreesOffset = 180},
		{sprite = "CRB_ChallengeTrackerSprites:sprChallengeStatusNormal", bgcolor = "white", degreesOffset = 90},
		{sprite = "CRB_ChallengeTrackerSprites:sprChallengeStatusPass", bgcolor = "white", degreesOffset = 90},
		{sprite = "CRB_ChallengeTrackerSprites:sprChallengeStatusFail", bgcolor = "white", degreesOffset = 90},
		{sprite = "ClientSprites:MiniMapObjectEdge", bgcolor = "white", degreesOffset = 0},
		{sprite = "ClientSprites:MiniMapPlayerArrow", bgcolor = "white", degreesOffset = 0},
		{sprite = "PlayerPathContent_TEMP:btn_PathListItemCloseFlyby", bgcolor = "white", degreesOffset = 90}
	}
	
    return o
end

function GroupRadar:Init()
    Apollo.RegisterAddon(self, true)
end

-----------------------------------------------------------------------------------------------
-- GroupRadar Default System Functions
-----------------------------------------------------------------------------------------------
function GroupRadar:OnLoad()
    -- Register handlers for events, slash commands, timer, etc.
	Apollo.RegisterSlashCommand("gradar", "OnSlashCommand", self)
	Apollo.RegisterSlashCommand("groupradar", "OnSlashCommand", self)
	
	-- TODO register these events only when GroupRadar is active
	Apollo.RegisterEventHandler("NextFrame", "OnUpdate", self)
	Apollo.RegisterEventHandler("Group_UpdatePosition", "OnGroupUpdatePosition", self)
	
	-- OptionsForm: loaded when needed
	self.wndOptions = nil
		
	-- general settings: on, off, testing
	self.bIsGroupRadarEnabled = true
	self.bIsTestEnabled = false
		
	-- rendering options
	self.bDisplayNearestOnly = false
	self.bDisplayFarthestOnly = false
	self.nMinDistance = 0
	self.nMaxDistance = 0
	self.nOverheadDistance = 50
	self.bOverheadText = false
	
	self.nArrowPosition = 50
	self.nArrowOpacity = 0.75
	self.nTextOpacity = 0.75
	self.nArrowStyle = 2
	
	self.aTrackedPartyMembers = {}
	self.tLastGroupPositions = {}
	
end


function GroupRadar:OnSave(eLevel) 
	-- TODO implement option to save accountwide or per character
	if eLevel ~= GameLib.CodeEnumAddonSaveLevel.Character then return end 
	
	return {
		-- general settings: on, off, testing
		groupRadarEnabled = self.bIsGroupRadarEnabled,
		testEnabled = self.bIsTestEnabled,
		
		-- rendering options
		displayNearestOnly = self.bDisplayNearestOnly,
		displayFarthestOnly = self.bDisplayFarthestOnly,
		minDistance = self.nMinDistance,
		maxDistance = self.nMaxDistance,
		overheadDistance = self.nOverheadDistance,
		bOverheadText = self.bOverheadText,
		arrowPosition = self.nArrowPosition,
		arrowOpacity = self.nArrowOpacity,
		textOpacity = self.nTextOpacity,
		arrowStyle = self.nArrowStyle
	}

end 


function GroupRadar:OnRestore(eLevel, tData)
	-- general settings: on, off, testing
	self.bIsGroupRadarEnabled = self:SetParamOnRestore(tData.groupRadarEnabled, self.bIsGroupRadarEnabled)
	self.bIsTestEnabled = self:SetParamOnRestore(tData.testEnabled, self.bIsTestEnabled)
		
	-- rendering options
	self.bDisplayNearestOnly = self:SetParamOnRestore(tData.displayNearestOnly, self.bDisplayNearestOnly)
	self.bDisplayFarthestOnly = self:SetParamOnRestore(tData.displayFarthestOnly, self.bDisplayFarthestOnly)
	self.nMinDistance = self:SetParamOnRestore(tData.minDistance, self.nMinDistance)
	self.nMaxDistance = self:SetParamOnRestore(tData.maxDistance, self.nMaxDistance)
	self.nOverheadDistance = self:SetParamOnRestore(tData.overheadDistance, self.nOverheadDistance)
	self.bOverheadText = self:SetParamOnRestore(tData.bOverheadText, self.bOverheadText)
	self.nArrowPosition = self:SetParamOnRestore(tData.arrowPosition, self.nArrowPosition)
	self.nArrowOpacity = self:SetParamOnRestore(tData.arrowOpacity, self.nArrowOpacity)
	self.nTextOpacity = self:SetParamOnRestore(tData.textOpacity, self.nTextOpacity)
	self.nArrowStyle = self:SetParamOnRestore(tData.arrowStyle, self.nArrowStyle)
end 


function GroupRadar:OnConfigure()
	
	if self.wndOptions == nil then
		self.wndOptions = Apollo.LoadForm("GroupRadar.xml", "OptionsForm", nil, self)
	end

	-- set active status-checkbox
	if self.bIsGroupRadarEnabled then
		self.wndOptions:FindChild("StatusOn"):SetCheck(true)
	elseif self.bIsTestEnabled then
		self.wndOptions:FindChild("StatusTest"):SetCheck(true)	
	else
		self.wndOptions:FindChild("StatusOff"):SetCheck(true)
	end
	
	-- set active mode-checkbox
	if self.bDisplayNearestOnly then
		self.wndOptions:FindChild("ModeNearest"):SetCheck(true)
	elseif self.bDisplayFarthestOnly then
		self.wndOptions:FindChild("ModeFarthest"):SetCheck(true)	
	else
		self.wndOptions:FindChild("ModeAll"):SetCheck(true)
	end
	
	-- set min distance
	self.wndOptions:FindChild("EditBoxMinDistance"):SetText(self.nMinDistance)
	self.wndOptions:FindChild("SliderMinDistance"):SetValue(self.nMinDistance)
	
	-- set overhead distance
	self.wndOptions:FindChild("EditBoxOverheadDistance"):SetText(self.nOverheadDistance)
	self.wndOptions:FindChild("SliderOverheadDistance"):SetValue(self.nOverheadDistance)

	-- set arrow position
	self.wndOptions:FindChild("EditBoxArrowPosition"):SetText(self.nArrowPosition)
	self.wndOptions:FindChild("SliderArrowPosition"):SetValue(self.nArrowPosition)
	
	-- set opacity values
	local nArrowOpacity = math.floor(self.nArrowOpacity * 100)
	self.wndOptions:FindChild("EditBoxArrowOpacity"):SetText(nArrowOpacity)
	self.wndOptions:FindChild("SliderArrowOpacity"):SetValue(nArrowOpacity)
	local nTextOpacity = math.floor(self.nTextOpacity * 100)
	self.wndOptions:FindChild("EditBoxTextOpacity"):SetText(nTextOpacity)
	self.wndOptions:FindChild("SliderTextOpacity"):SetValue(nTextOpacity)

	-- set active style-checkbox
	self.wndOptions:FindChild("Style"..self.nArrowStyle):SetCheck(true)
	
	self.wndOptions:Show(true)
end

-----------------------------------------------------------------------------------------------
-- GroupRadar Registered Functions
-----------------------------------------------------------------------------------------------

-- on SlashCommand "/gradar"
function GroupRadar:OnSlashCommand(strCommand, strArg)
	
	local mixedArgs = self:explodeString(strArg, ' ')
	if mixedArgs[2] then
		local nTemp = tonumber(mixedArgs[2])
		if nTemp then
			if nTemp < 0 then
				nTemp = 0
			end
		end
		mixedArgs[2] = nTemp
	end
	
	if mixedArgs[1] == "show" then 
		Print("GroupRadar enabled - use 'gradar hide' to disable")
		self:Enable()
	elseif mixedArgs[1] == "hide" then 
		Print("GroupRadar disabled - use 'gradar show' to enable")
		self:Disable()
	elseif mixedArgs[1] == "toggle" then 
		if self.bIsGroupRadarEnabled then
			self:Disable()
			Print("GroupRadar toggled \"off\"")
		else
			self:Enable()
			Print("GroupRadar toggled \"on\"")
		end
	elseif mixedArgs[1] == "showohtxt" then 
		Print("GroupRadar: overhead-texts enabled - use 'gradar hideohtxt' to disable")
		self:EnableOverheadTexts()
	elseif mixedArgs[1] == "hideohtxt" then 
		Print("GroupRadar overhead-texts disabled - use 'gradar showohtxt' to enable")
		self:DisableOverheadTexts()
	elseif mixedArgs[1] == "toggleohtxt" then 
		if self.bOverheadText == true then
			self:DisableOverheadTexts()
			Print("GroupRadar overhead-texts toggled \"off\"")
		else
			self:EnableOverheadTexts()
			Print("GroupRadar overhead-texts toggled \"on\"")
		end
	elseif mixedArgs[1] == "maxdistance" then
		if mixedArgs[2] then
			self:SetMaxDistance(mixedArgs[2])
			if mixedArgs[2] == 0 then
				Print("Maximal tracking distance disabled!")
			else
				Print("Maximal tracking distance set to " .. mixedArgs[2] .. "! Disable with: \"/groupradar maxdistance 0\"")
			end
		end
	else
		self:OnConfigure()
	end
end


-- on "NextFrame"
function GroupRadar:OnUpdate()
	if self.bIsGroupRadarEnabled or self.bIsTestEnabled then
		
		local nCurrentTime = GameLib.GetGameTime()
		local nElapsedTime = nCurrentTime - nLastTime
		
		nLastTime = nCurrentTime
		nUpdateTime = nUpdateTime + nElapsedTime
		
		if bUpdateForced or (nUpdateTime >= UPDATETHROTTLE) then
			self:UpdateArrows()
			self:RefreshArrows()
			nUpdateTime, nRefreshTime = 0, 0
			bUpdateForced = false
		else
			nRefreshTime = nRefreshTime + nElapsedTime
			if nRefreshTime >= REFRESHTHROTTLE then 
				self:RefreshArrows()
				nRefreshTime = 0
			end
		end
	end
end


-- on "Group_UpdatePosition"
function GroupRadar:OnGroupUpdatePosition(tPositions)
	
	for idx,tRecord in pairs(tPositions) do 
		self.tLastGroupPositions[tRecord.nIndex] = tRecord
	end
end

-----------------------------------------------------------------------------------------------
-- Arrow Functions
-----------------------------------------------------------------------------------------------

-- UpdateArrows:
-- Reads GroupData and creates arrows.
function GroupRadar:UpdateArrows()
	
	-- GroupRadar is active
	-- => Read GroupData
	-- => Decide which Arrows to draw
	-- => Create Arrows
	if self.bIsGroupRadarEnabled and GroupLib.InGroup() and not GroupLib.InRaid() then
		
		local unitPlayer = GameLib.GetPlayerUnit()
		-- prevent nullpointerexception during ui(re)load while the playerUnit doesn't exist
		if not unitPlayer then
			return
		end
	
		local aPartyMembers = {}
		local nDistance = nil
		local nDistanceIndex = nil
		local nCurrentContinentId = nil
		local tCurrentZoneMap = GameLib.GetCurrentZoneMap()
		if tCurrentZoneMap then
			nCurrentContinentId = tCurrentZoneMap.continentId
		end

		-- try to get units for all groupmembers, add all groupmembers in range				
		for nGroupMemberIndex = 1, 5, 1 do			
			
			local unitCurrentMember = GroupLib.GetUnitForGroupMember(nGroupMemberIndex)
			
			if unitCurrentMember ~= nil then
				
				-- skip ourselves
				if not unitCurrentMember:IsThePlayer() then
					local nDistanceToPlayer = self:CalculateDistanceBetweenPositions2D(unitPlayer:GetPosition(), unitCurrentMember:GetPosition())

					-- save index of nearest/farthest player if in "displayNearest/FarthestOnly"-mode
					if 
						nDistanceToPlayer and (
							(self.bDisplayNearestOnly and ((nDistance == nil) or (nDistance ~= nil and nDistance > nDistanceToPlayer)))
							or (self.bDisplayFarthestOnly and ((nDistance == nil) or (nDistance ~= nil and nDistance < nDistanceToPlayer)))
						)
					then
						nDistance = nDistanceToPlayer
						nDistanceIndex = nGroupMemberIndex
					end
					
					aPartyMembers[nGroupMemberIndex] = {name = unitCurrentMember:GetName(), unit = unitCurrentMember, distance = nDistanceToPlayer, role = nil, index = nGroupMemberIndex} -- TODO getRole
				end
			
			-- try to use data provided by "Group_UpdatePosition"
			else
				
				local tGroupMember = GroupLib.GetGroupMember(nGroupMemberIndex)

				if tGroupMember and self.tLastGroupPositions and self.tLastGroupPositions[nGroupMemberIndex] and nCurrentContinentId == self.tLastGroupPositions[nGroupMemberIndex].tZoneMap.continentId then
					local nDistanceToPlayer = self:CalculateDistanceBetweenPositions2D(unitPlayer:GetPosition(), self.tLastGroupPositions[nGroupMemberIndex].tWorldLoc)
					-- save index of nearest/farthest player if in "displayNearest/FarthestOnly"-mode
					if 
						nDistanceToPlayer and (
							(self.bDisplayNearestOnly and ((nDistance == nil) or (nDistance ~= nil and nDistance > nDistanceToPlayer)))
							or (self.bDisplayFarthestOnly and ((nDistance == nil) or (nDistance ~= nil and nDistance < nDistanceToPlayer)))
						)
					then
						nDistance = nDistanceToPlayer
						nDistanceIndex = nGroupMemberIndex
					end
					
					aPartyMembers[nGroupMemberIndex] = {name = tGroupMember.strCharacterName, unit = nil, distance = nDistanceToPlayer, role = nil, index = nGroupMemberIndex} -- TODO getRole
				end
			end	
		end

		-- loop over all found groupmembers and create or remove arrows if needed
		for nGroupMemberIndex = 1, 5, 1 do
			if
				aPartyMembers[nGroupMemberIndex] -- check if we have a groupmember
				and aPartyMembers[nGroupMemberIndex].distance and aPartyMembers[nGroupMemberIndex].distance > self.nMinDistance -- check minDistance
				and (self.nMaxDistance == 0 or (self.nMaxDistance ~= 0 and aPartyMembers[nGroupMemberIndex].distance < self.nMaxDistance)) -- check maxDistance
				and (nDistanceIndex == nil or (nDistanceIndex ~= nil and nDistanceIndex == nGroupMemberIndex)) 	-- check displaymode (all/nearest/farthest)
			then
				if not self.aTrackedPartyMembers[nGroupMemberIndex] then
					self.aTrackedPartyMembers[nGroupMemberIndex] = {
						name = nil,
						arrow = Apollo.LoadForm("GroupRadar.xml", "NavArrowForm", nil, self),
						arrowOverhead = Apollo.LoadForm("GroupRadar.xml", "NavArrowFormOverhead", "InWorldHudStratum", self)
					}
					self:SetDisplayOptions(self.aTrackedPartyMembers[nGroupMemberIndex].arrow, false)
					self:SetDisplayOptions(self.aTrackedPartyMembers[nGroupMemberIndex].arrowOverhead, true)
				end
				self.aTrackedPartyMembers[nGroupMemberIndex].name = aPartyMembers[nGroupMemberIndex].name
				self.aTrackedPartyMembers[nGroupMemberIndex].index = aPartyMembers[nGroupMemberIndex].index
				-- TODO add role-colors
				self.aTrackedPartyMembers[nGroupMemberIndex].arrow:FindChild("Name"):SetAML("<P Align=\"Center\" Font=\"CRB_InterfaceSmall_O\">" .. aPartyMembers[nGroupMemberIndex].name .. "</P>")
				self.aTrackedPartyMembers[nGroupMemberIndex].arrowOverhead:FindChild("Name"):SetAML("<P Align=\"Center\" Font=\"CRB_InterfaceSmall_O\">" .. aPartyMembers[nGroupMemberIndex].name .. "</P>")

				
			elseif self.aTrackedPartyMembers[nGroupMemberIndex] ~= nil then
				-- remove arrows
				self.aTrackedPartyMembers[nGroupMemberIndex].arrow:Show(false)
				self.aTrackedPartyMembers[nGroupMemberIndex].arrowOverhead:Show(false)
				self.aTrackedPartyMembers[nGroupMemberIndex] = nil
			end
		end
	
	
	-- GroupRadar is in testing Mode
	-- => use positions relative to the players position to render the arrows
	elseif self.bIsTestEnabled then
		
		for nGroupMemberIndex = 0, 2, 1 do
			
			if not self.aTrackedPartyMembers[nGroupMemberIndex] then
				self.aTrackedPartyMembers[nGroupMemberIndex] = {
					name = "GroupRadar - Test"..nGroupMemberIndex,
					arrow = Apollo.LoadForm("GroupRadar.xml", "NavArrowForm", nil, self),
					arrowOverhead = Apollo.LoadForm("GroupRadar.xml", "NavArrowFormOverhead", "InWorldHudStratum", self)
				}
				self:SetDisplayOptions(self.aTrackedPartyMembers[nGroupMemberIndex].arrow, false)
				self:SetDisplayOptions(self.aTrackedPartyMembers[nGroupMemberIndex].arrowOverhead, true)
			end
		end

	-- GroupRadar is disabled
	-- => clear tracked data
	else
		self:ResetArrows()
	end

end


-- RefreshArrows:
-- Refreshes arrows created by "UpdateArrows".
function GroupRadar:RefreshArrows()

	local unitPlayer = GameLib.GetPlayerUnit()
	-- prevent nullpointerexception during ui(re)load while the playerUnit doesn't exist
	if not unitPlayer then
		return
	end
	
	-- if GroupRadar is enabled ..
	if self.bIsGroupRadarEnabled then
		
		-- this check is needed because in some situations currentZoneMap is nil
		local nCurrentContinentId = nil
		local tCurrentZoneMap = GameLib.GetCurrentZoneMap()
		if tCurrentZoneMap then
			nCurrentContinentId = tCurrentZoneMap.continentId
		end

		-- .. loop over tracked players
		for idx,tRecord in pairs(self.aTrackedPartyMembers) do
			-- we have to verify that the unit is stil accessible to prevent client-crashes
			local unitCurrentMember = GroupLib.GetUnitForGroupMember(idx);

			-- tracked player is in "unit-rendering-range"
			if unitCurrentMember ~= nil then
				self:RedrawArrow(unitPlayer, unitCurrentMember:GetPosition(), tRecord.name, tRecord.arrow, tRecord.arrowOverhead, unitCurrentMember)
			
			-- tracked player is out of "unit-rendering-range" but we have the players position
			elseif self.tLastGroupPositions[idx] and nCurrentContinentId == self.tLastGroupPositions[idx].tZoneMap.continentId then
				self:RedrawArrow(unitPlayer, self.tLastGroupPositions[idx].tWorldLoc, tRecord.name, tRecord.arrow, tRecord.arrowOverhead, nil)
			end
			
		end
	
	-- if GroupRadar is in testing mode ..
	elseif self.bIsTestEnabled then

		for idx, tRecord in pairs(self.aTrackedPartyMembers) do 
			local tTrackedPlayerPosition = unitPlayer:GetPosition()
			-- prevent nullpointerexception during ui(re)load while the playerUnit has no position
			if tTrackedPlayerPosition then
				tTrackedPlayerPosition.z = tTrackedPlayerPosition.z + 50 - 25 * idx
				tTrackedPlayerPosition.x = tTrackedPlayerPosition.x + 25 * idx
				self:RedrawArrow(unitPlayer, tTrackedPlayerPosition, tRecord.name, tRecord.arrow, tRecord.arrowOverhead, unitPlayer)
			end
		end 
	
	end

end


-- RedrawArrow:
-- Calculates new arrow-properties and redraws the arrow accordingly.
function GroupRadar:RedrawArrow(unitPlayer, tTrackedPlayerPosition, sTrackedPlayerName, oArrow, oArrowOverhead, unitTrackedPlayer )

	if not oArrow or (not unitTrackedPlayer and not tTrackedPlayerPosition) then
		return
	end

	local tPlayerPosition = unitPlayer:GetPosition()
	local nDistance, nAngle = self:GetVectorBetweenPositions2D(tPlayerPosition, tTrackedPlayerPosition)
	-- force update if drawing an arrow with distance < minDistance
	if nDistance < self.nMinDistance or (self.nMaxDistance ~= 0 and nDistance > self.nMaxDistance) then
		bUpdateForced = true
	end
	
	-- Determine if we need to plant the arrow over the tracked player's head
	local bShowOverheadArrow = false
	if unitTrackedPlayer and nDistance < self.nOverheadDistance then
		bShowOverheadArrow = GameLib.GetUnitScreenPosition(unitTrackedPlayer).bOnScreen
	end
	
	if bShowOverheadArrow then
		-- attach arrow to tracked playermodel or its mount
		self:SetArrowUnit(unitTrackedPlayer, oArrowOverhead)
		
		if self.bOverheadText then
			oArrowOverhead:FindChild("Distance"):SetText(string.format("%.2f", nDistance).."m")
			oArrowOverhead:FindChild("Name"):SetAML("<P Align=\"Center\" Font=\"CRB_InterfaceSmall_O\">" .. sTrackedPlayerName .. "</P>")
		end
		
		if not oArrowOverhead:IsShown() or oArrow:IsShown() then
			oArrow:Show(false)
			if not self.bOverheadText then 
				oArrowOverhead:FindChild("Name"):Show(false)
				oArrowOverhead:FindChild("Distance"):Show(false)
			end
			oArrowOverhead:Show(true)
		end
	else
		-- calculate arrow rotation
		local nPlayerFacingAngle = unitPlayer:GetHeading()
		if nPlayerFacingAngle < 0 then
			nPlayerFacingAngle = nPlayerFacingAngle * -1
		else
			nPlayerFacingAngle = 2 * math.pi - nPlayerFacingAngle
		end

		local nArrowFacingAngle = nAngle - nPlayerFacingAngle

		-- calculate new window-position
		local nPositionHorizontal = math.sin(nArrowFacingAngle) * self.nArrowPosition * 4
		local nPositionVertical = -math.cos(nArrowFacingAngle) * self.nArrowPosition * 4
		local nOffsetRigt = nPositionHorizontal + 75
		local nOffsetLeft = nPositionHorizontal - 75
		local nOffsetTop = nPositionVertical
		local nOffsetBottom = nPositionVertical + 150
		
		-- set rotation and position
		oArrow:FindChild("Arrow"):SetRotation((nArrowFacingAngle * (180 / math.pi)) + self.aArrowStyles[self.nArrowStyle].degreesOffset)
		oArrow:SetAnchorOffsets(nOffsetLeft, nOffsetTop, nOffsetRigt, nOffsetBottom)
		
		oArrow:FindChild("Distance"):SetText(string.format("%.2f", nDistance).."m")
		oArrow:FindChild("Name"):SetAML("<P Align=\"Center\" Font=\"CRB_InterfaceSmall_O\">" .. sTrackedPlayerName .. "</P>")
		
		if not oArrow:IsShown() or oArrowOverhead:IsShown() then
			oArrow:Show(true)
			oArrowOverhead:Show(false)
		end
	end

end

-- SetArrowUnit:
-- Sets the unit the arrow should be displayed on. Includes mount-fix.
function GroupRadar:SetArrowUnit(unitPlayer, oArrow)
	if unitPlayer:IsMounted() and oArrow:GetUnit() ~= unitPlayer:GetUnitMount() then
		oArrow:SetUnit(unitPlayer:GetUnitMount(), 1)
	elseif not unitPlayer:IsMounted() and oArrow:GetUnit() ~= unitPlayer then
		oArrow:SetUnit(unitPlayer, 1)
	end
end

-- ResetArrows:
-- Clears all arrows and tracked-player-data.
function GroupRadar:ResetArrows()
	if self.aTrackedPartyMembers then
		for idx, tTrackedPartyMember in pairs(self.aTrackedPartyMembers) do 
			tTrackedPartyMember.arrow:Show(false)
			tTrackedPartyMember.arrowOverhead:Show(false)
		end
		self.aTrackedPartyMembers = {}
	end
end

-- ApplyDisplayOptions:
-- updates all existing arrows with current options
function GroupRadar:ApplyDisplayOptions()
	if self.aTrackedPartyMembers then	
		for idx, tTrackedPartyMember in pairs(self.aTrackedPartyMembers) do 
			self:SetDisplayOptions(tTrackedPartyMember.arrow, false)
			self:SetDisplayOptions(tTrackedPartyMember.arrowOverhead, true)
		end
		bUpdateForced = true
	end
end

-- SetDisplayOptions:
-- updates given arrow with current options
function GroupRadar:SetDisplayOptions(oArrow, blOverhead)
	oArrow:FindChild("Arrow"):SetOpacity(self.nArrowOpacity)
	oArrow:FindChild("Arrow"):SetSprite(self.aArrowStyles[self.nArrowStyle].sprite)
	oArrow:FindChild("Arrow"):SetBGColor(self.aArrowStyles[self.nArrowStyle].bgcolor)
	oArrow:FindChild("Distance"):SetOpacity(self.nTextOpacity)
	oArrow:FindChild("Name"):SetOpacity(self.nTextOpacity)
	if blOverhead == true then
		oArrow:FindChild("Arrow"):SetRotation(180 + self.aArrowStyles[self.nArrowStyle].degreesOffset)
		oArrow:SetAnchorOffsets(nDefaultOffsetLeft, nDefaultOffsetTop, nDefaultOffsetRight, nDefaultOffsetBottom )
	end
end


-----------------------------------------------------------------------------------------------
-- Distance and Angle Functions
-----------------------------------------------------------------------------------------------
function GroupRadar:CalculateAngle(nX, nZ)
	return math.atan2(nX, -nZ)
end

function GroupRadar:CalculateDistanceAndXZDeltaBetweenPositions2D(tPosition1, tPosition2)
	if not tPosition1 or not tPosition2 then
		return
	end
	
	local nDeltaX = tPosition2.x - tPosition1.x
	local nDeltaZ = tPosition2.z - tPosition1.z
	return math.sqrt(math.pow(nDeltaX, 2) + math.pow(nDeltaZ, 2)), nDeltaX, nDeltaZ
end

function GroupRadar:CalculateDistanceBetweenPositions2D(tPosition1, tPosition2)
	if not tPosition1 or not tPosition2 then
		return
	end
	
	local nDistance, nDeltaX, nDeltaZ = self:CalculateDistanceAndXZDeltaBetweenPositions2D(tPosition1, tPosition2)
	return nDistance
end
	
function GroupRadar:GetVectorBetweenPositions2D(tPosition1, tPosition2)	
	if not tPosition1 or not tPosition2 then
		return
	end

	local nDistance, nDeltaX, nDeltaZ = self:CalculateDistanceAndXZDeltaBetweenPositions2D(tPosition1, tPosition2)
	return nDistance, self:CalculateAngle(nDeltaX, nDeltaZ)			
end


---------------------------------------------------------------------------------------------------
-- OptionsForm Functions
---------------------------------------------------------------------------------------------------

function GroupRadar:OnClose( wndHandler, wndControl, eMouseButton )
	if self.wndOptions ~= nil then
		self.wndOptions:Show(false)
		self.wndOptions = nil
	end
end


-- status
function GroupRadar:OnStatusChange( wndHandler, wndControl, eMouseButton )
	local sName = wndControl:GetName()
	if sName == "StatusOn" then
		self:Enable()
	elseif sName == "StatusTest" then
		self:EnableTest()
	else
		self:Disable()
	end		
end


-- mode
function GroupRadar:OnModeChange( wndHandler, wndControl, eMouseButton )
	local sName = wndControl:GetName()
	if sName == "ModeNearest" then
		self:SetModeNearest()
	elseif sName == "ModeFarthest" then
		self:SetModeFarthest()
	else
		self:SetModeAll()
	end
end


-- TODO: unify methods for slider handling. use naming-conventions to controll behavior
-- arrowOpacity
function GroupRadar:OnArrowOpacityChanging( wndHandler, wndControl, fNewValue, fOldValue, bOkToChange )
	local nNewValue = math.floor(fNewValue)

	self.wndOptions:FindChild("EditBoxArrowOpacity"):SetText(nNewValue)
	self:SetArrowOpacity(nNewValue)
	
	wndControl:SetValue(fNewValue)
end

function GroupRadar:OnEditArrowOpacity( wndHandler, wndControl, strText )
	local nValue = tonumber(strText)
	if nValue then
		nValue = self:CheckValueBetween(nValue, 0, 100)
		wndControl:SetText(nValue)
		self.wndOptions:FindChild("SliderArrowOpacity"):SetValue(nValue)
		self:SetArrowOpacity(nValue)
	else
		wndControl:SetText(math.floor(self.nArrowOpacity*100))
	end
end

function GroupRadar:OnEscapeArrowOpacity( wndHandler, wndControl )
	wndControl:SetText(math.floor(self.nArrowOpacity*100))
end


-- arrowPosition
function GroupRadar:OnArrowPositionChanging( wndHandler, wndControl, fNewValue, fOldValue, bOkToChange )
	local nNewValue = math.floor(fNewValue)

	self.wndOptions:FindChild("EditBoxArrowPosition"):SetText(nNewValue)
	self:SetArrowPosition(nNewValue)
	
	wndControl:SetValue(fNewValue)
end

function GroupRadar:OnEditArrowPosition( wndHandler, wndControl, strText )
	local nValue = tonumber(strText)
	if nValue then
		nValue = self:CheckValueBetween(nValue, 0, 100)
		wndControl:SetText(nValue)
		self.wndOptions:FindChild("SliderArrowPosition"):SetValue(nValue)
		self:SetArrowPosition(nValue)
	else
		wndControl:SetText(math.floor(self.nArrowPosition))
	end
end

function GroupRadar:OnEscapeArrowPosition( wndHandler, wndControl )
	wndControl:SetText(self.nArrowPosition)
end


-- textOpacity
function GroupRadar:OnTextOpacityChanging( wndHandler, wndControl, fNewValue, fOldValue, bOkToChange )
	local nNewValue = math.floor(fNewValue)

	self.wndOptions:FindChild("EditBoxTextOpacity"):SetText(nNewValue)
	self:SetTextOpacity(nNewValue)
	
	wndControl:SetValue(fNewValue)
end

function GroupRadar:OnEditTextOpacity( wndHandler, wndControl, strText )
	local nValue = tonumber(strText)
	if nValue then
		nValue = self:CheckValueBetween(nValue, 0, 100)
		wndControl:SetText(nValue)
		self.wndOptions:FindChild("SliderTextOpacity"):SetValue(nValue)
		self:SetTextOpacity(nValue)
	else
		wndControl:SetText(math.floor(self.nTextOpacity*100))
	end
end

function GroupRadar:OnEscapeTextOpacity( wndHandler, wndControl )
	wndControl:SetText(math.floor(self.nTextOpacity*100))
end


-- minDistance
function GroupRadar:OnMinDistanceChanging( wndHandler, wndControl, fNewValue, fOldValue, bOkToChange )
	local nNewValue = math.floor(fNewValue)

	self.wndOptions:FindChild("EditBoxMinDistance"):SetText(nNewValue)
	self:SetMinDistance(nNewValue)
	
	wndControl:SetValue(fNewValue)
end

function GroupRadar:OnEditMinDistance( wndHandler, wndControl, strText )
	local nValue = tonumber(strText)
	if nValue then
		nValue = self:CheckValueBetween(nValue, 0, 100)
		wndControl:SetText(nValue)
		self.wndOptions:FindChild("SliderMinDistance"):SetValue(nValue)
		self:SetMinDistance(nValue)
	else
		wndControl:SetText(math.floor(self.nMinDistance))
	end
end

function GroupRadar:OnEscapeMinDistance( wndHandler, wndControl )
	wndControl:SetText(self.nMinDistance)
end


-- overheadDistance
function GroupRadar:OnOverheadDistanceChanging( wndHandler, wndControl, fNewValue, fOldValue, bOkToChange )
	local nNewValue = math.floor(fNewValue)

	self.wndOptions:FindChild("EditBoxOverheadDistance"):SetText(nNewValue)
	self:SetOverheadDistance(nNewValue)
	
	wndControl:SetValue(fNewValue)
end

function GroupRadar:OnEditOverheadDistance( wndHandler, wndControl, strText )
	local nValue = tonumber(strText)
	if nValue then
		nValue = self:CheckValueBetween(nValue, 0, 100)
		wndControl:SetText(nValue)
		self.wndOptions:FindChild("SliderOverheadDistance"):SetValue(nValue)
		self:SetOverheadDistance(nValue)
	else
		wndControl:SetText(math.floor(self.nOverheadDistance))
	end
end

function GroupRadar:OnEscapeOverheadDistance( wndHandler, wndControl )
	wndControl:SetText(self.nOverheadDistance)
end


-- style
function GroupRadar:OnStyleChange( wndHandler, wndControl, eMouseButton )
	self:SetDisplayStyle(string.sub(wndControl:GetName(), 6))
end


-----------------------------------------------------------------------------------------------
-- OptionsSetter Functions
-----------------------------------------------------------------------------------------------

function GroupRadar:Enable()
	self.bIsGroupRadarEnabled = true
	self.bIsTestEnabled = false
	self:ResetArrows()
	bUpdateForced = true
end

function GroupRadar:EnableTest()
	self.bIsGroupRadarEnabled = false
	self.bIsTestEnabled = true
	self:ResetArrows()
	bUpdateForced = true
end

function GroupRadar:Disable()
	self.bIsGroupRadarEnabled = false
	self.bIsTestEnabled = false
	self:ResetArrows()
	bUpdateForced = true
end

function GroupRadar:EnableOverheadTexts()
	self.bOverheadText = true
	self:ResetArrows()
	bUpdateForced = true
end

function GroupRadar:DisableOverheadTexts()
	self.bOverheadText = false
	self:ResetArrows()
	bUpdateForced = true
end

function GroupRadar:SetArrowPosition(nOffset)
	self.nArrowPosition = nOffset
	self:ApplyDisplayOptions()
end

function GroupRadar:SetArrowOpacity(nOpacity)
	self.nArrowOpacity = nOpacity / 100
	self:ApplyDisplayOptions()
end

function GroupRadar:SetTextOpacity(nOpacity)
	self.nTextOpacity = nOpacity / 100
	self:ApplyDisplayOptions()
end

function GroupRadar:SetModeAll()
	self.bDisplayNearestOnly = false
	self.bDisplayFarthestOnly = false
	bUpdateForced = true
end

function GroupRadar:SetModeNearest()
	self.bDisplayNearestOnly = true
	self.bDisplayFarthestOnly = false
	bUpdateForced = true
end

function GroupRadar:SetModeFarthest()
	self.bDisplayNearestOnly = false
	self.bDisplayFarthestOnly = true
	bUpdateForced = true
end

function GroupRadar:SetDisplayStyle(nStyleID)
	self.nArrowStyle = tonumber(nStyleID)
	self:ApplyDisplayOptions()
end

function GroupRadar:SetMinDistance(nMinDistance)
	self.nMinDistance = nMinDistance
end

function GroupRadar:SetMaxDistance(nMaxDistance)
	self.nMaxDistance = nMaxDistance
end

function GroupRadar:SetOverheadDistance(nOverheadDistance)
	self.nOverheadDistance = nOverheadDistance
end


-----------------------------------------------------------------------------------------------
-- Helper Functions
-----------------------------------------------------------------------------------------------
function GroupRadar:SetParamOnRestore(mixedLoadedValue, mixedDefaultValue)
	if mixedLoadedValue ~= nil then
		return mixedLoadedValue
	end
	return mixedDefaultValue
end

function GroupRadar:CheckValueBetween(nValue, nFrom, nTo)
	if nValue < nFrom then
		return nFrom
	elseif nValue > nTo then
		return nTo
	end
	return nValue
end

function GroupRadar:explodeString(sString, sDivider)
    if (sDivider == '') then
		return false
	end
    local pos , arr = 0, {}
    for st, sp in function() return string.find(sString, sDivider, pos, true) end do
        table.insert(arr, string.sub(sString, pos, st-1))
        pos = sp + 1
    end
    table.insert(arr, string.sub(sString, pos))
    return arr
end

-----------------------------------------------------------------------------------------------
-- GroupRadar Instance
-----------------------------------------------------------------------------------------------
local GroupRadarInst = GroupRadar:new()
GroupRadarInst:Init()
